@@ -27,6 +27,7 @@ from typing import (
     AsyncGenerator,
     Awaitable,
     Callable,
+    Generator,
     List,
     Mapping,
     Optional,
@@ -36,17 +37,22 @@ from typing import (
 )
 
 import zarr
-from anyio import start_blocking_portal
 from zarr._storage.store import StoreV3
 from zarr.util import normalize_storage_path
 
+from arraylake_client.async_utils import (
+    BlockingPortal,
+    BlockingPortalGenerator,
+    cached_portal,
+)
 from arraylake_client.chunkstore import Chunkstore
-from arraylake_client.commits import CommitData, CommitLog, CommitTree
+from arraylake_client.commits import CommitData, CommitLog
 from arraylake_client.exceptions import (
     CommitFailedError,
     DocumentNotFoundError,
     InvalidPrefixError,
 )
+from arraylake_client.log_util import get_logger
 from arraylake_client.metastore import MetastoreDatabase
 from arraylake_client.types import (
     Author,
@@ -70,6 +76,8 @@ from arraylake_client.zarr_util import (
     meta_root,
 )
 
+logger = get_logger(__name__)
+
 metadata_collection = CollectionName("metadata")
 chunks_collection = CollectionName("chunks")
 
@@ -175,6 +183,14 @@ class AsyncRepo:
         await self.chunkstore.__aexit__(*args, **kwargs)
         self._OPEN = False
 
+    def __getstate__(self):
+        return self.db, self.chunkstore, self.repo_name, self.author, self._session, self._OPEN, self._writable
+
+    def __setstate__(self, state):
+        self.db, self.chunkstore, self.repo_name, self.author, self._session, self._OPEN, self._writable = state
+        # commit_data can be large and is not needed for most operations, therefore we omit it during serialization.
+        self._commit_data = None
+
     def __repr__(self):
         repo_name = self.repo_name
         status = "OPEN" if self._OPEN else "CLOSED"
@@ -187,17 +203,19 @@ class AsyncRepo:
             raise ValueError("There is no session active. You have to call checkout first.")
         return self._session
 
-    @property
-    def commit_data(self) -> CommitData:
+    async def commit_data(self) -> CommitData:
+        """Returns the #CommitData for the current session."""
         if self._commit_data is None:
-            raise ValueError("You check out a branch to see the commit log.")
+            if not self._OPEN:
+                raise ValueError("must be in async context to fetch commit_data")
+            # lazily refresh commit data
+            await self._refresh_commit_data()
         return self._commit_data
 
-    @property
-    def commit_log(self) -> CommitLog:
+    async def commit_log(self) -> CommitLog:
         """Returns the #CommitLog for the current session."""
 
-        return CommitLog(self.repo_name, self.session["base_commit"], self.commit_data)
+        return CommitLog(self.repo_name, self.session["base_commit"], await self.commit_data())
 
     async def status(self, limit: int = 1000) -> SessionStatus:
         """Returns the #SessionStatus for the current session.
@@ -235,8 +253,7 @@ class AsyncRepo:
         """
 
         await self._refresh_commit_data()
-        commit_data = self.commit_data
-        commit, branch = self.commit_data.get_ref(ref)
+        commit, branch = (await self.commit_data()).get_ref(ref)
 
         session_start_time = datetime.datetime.utcnow()
         # TODO: replace this with a pydantic model
@@ -248,7 +265,6 @@ class AsyncRepo:
         else:
             self._writable = True
 
-        self._commit_tree = CommitTree(commit, commit_data.commits) if commit else None
         return commit
 
     async def commit(self, message: str, auto_ff=True) -> CommitID:
@@ -295,7 +311,7 @@ class AsyncRepo:
             else:
                 raise err
 
-        new_branch = (self.session["branch"] not in self.commit_data.branches) or (self.session["base_commit"] is None)
+        new_branch = (self.session["branch"] not in (await self.commit_data()).branches) or (self.session["base_commit"] is None)
         try:
             await self.db.update_branch(
                 self.session["branch"], base_commit=self.session["base_commit"], new_commit=new_commit, new_branch=new_branch
@@ -328,7 +344,7 @@ class AsyncRepo:
 
         await self._refresh_commit_data()
         branch = self.session["branch"]
-        branch_latest_commit = self._commit_data.branches.get(branch, None)
+        branch_latest_commit = (await self.commit_data()).branches.get(branch, None)
         session_base_commit = self._session["base_commit"]
 
         if branch_latest_commit is None:
@@ -348,7 +364,7 @@ class AsyncRepo:
             return branch_latest_commit
 
         # this is different from self.commit_log because it starts from branch_latest_commit
-        commit_log = CommitLog(self.repo_name, branch_latest_commit, self.commit_data)
+        commit_log = CommitLog(self.repo_name, branch_latest_commit, await self.commit_data())
         upstream_modifications = set()
         for commit in commit_log:
             if commit.id == session_base_commit:
@@ -363,7 +379,7 @@ class AsyncRepo:
             raise RuntimeError(f"Can't fast-forward due to conflicting paths {conflicting_paths}.")
         return branch_latest_commit
 
-    def new_branch(self, branch_name: str) -> None:
+    async def new_branch(self, branch_name: str) -> None:
         """Create a new branch based on the current session reference
 
         Args:
@@ -374,7 +390,7 @@ class AsyncRepo:
         if self._session is None:
             # this was added to make mypy happy but is not covered by tests
             raise ValueError("There is no session active. You have to call checkout first.")
-        if branch in self.commit_data.branches or branch == self.session["branch"]:
+        if branch in (await self.commit_data()).branches or branch == self.session["branch"]:
             raise ValueError(f"Branch {branch} already exists.")
         self._session["branch"] = branch
 
@@ -720,13 +736,20 @@ class AsyncRepo:
             return
 
         start = len(path_query) + 1
+        groups_seen = set()
         async for path in self._list(path_query, all_subdirs=False):
+            # a group may end with .group.json (explicit) or have no
+            # suffix (implicit). in the group
+            # (in progress) should this logic be tweaked to find items without a suffix too (i.e a dir?)
             if path.endswith(".group.json"):
-                # the presence of a group.json indicates that there is a subdirectory
-                # NOTE: it is ambiguous whether we should append a trailing slash
-                # The spec suggests yes, but zarr-python doesn't seem to like this
-                yield path[start:-11]  # + "/"
-            yield path[start:]
+                p = path[start:-11]
+                if p not in groups_seen:
+                    yield path[start:]
+                    groups_seen.add(p)
+            else:
+                # trim the start of the path off, we only want what's in the dir
+                # e.g. meta/root/baz.array.json -> baz.array.json
+                yield path[start:]
 
     async def _list_prefix(self, prefix: str) -> AsyncGenerator[str, None]:
         """List a prefix in the metastore
@@ -863,6 +886,10 @@ class Repo:
     """Synchronous interface to Arraylake repo."""
 
     _arepo: AsyncRepo
+    _OPEN: bool
+    _cached_portal_context: Optional[BlockingPortalGenerator]
+    _portal: Optional[BlockingPortal]
+    _arepo_context: Optional[Generator]
 
     def __init__(self, arepo: AsyncRepo):
         """
@@ -872,7 +899,12 @@ class Repo:
             arepo: An existing AsyncRepo
         """
         self._arepo = arepo
+
+        # default / unset attributes used to wrap AsyncRepo
         self._OPEN = False
+        self._cached_portal_context = None
+        self._portal = None
+        self._arepo_context = None
 
     @classmethod
     def from_metastore_and_chunkstore(cls, metastore_db: MetastoreDatabase, chunkstore: Chunkstore, name: str, author: Author) -> Repo:
@@ -893,21 +925,23 @@ class Repo:
         return self._arepo.repo_name
 
     def _open(self):
-        self._blocking_portal = start_blocking_portal()
-        self._portal = self._blocking_portal.__enter__()
+        """using a AnyIO blocking portal, enter the AsyncRepo's async context"""
+
+        # Acquire the blocking portal .acquire() returns a context manager that when entered, provisions a BlockingPortal.
+        # and when exited, deprovisions the BlockingPortal (provided there are no other active references, e.g. from
+        # other repo objects).
+        self._cached_portal_context = cached_portal.acquire()
+        self._portal = self._cached_portal_context.__enter__()
+        # wrap_async_context_manager wraps the AsyncRepo object and .__enter__() enters the async context (synchronously)
         self._arepo_context = self._portal.wrap_async_context_manager(self._arepo)
         self._arepo_context.__enter__()
-        self._portal._check_running()
+
+        # _OPEN signals that the portal and async context are ready for use
         self._OPEN = True
 
     def _synchronize(self, method, *args, **kwargs):
         if not self._OPEN:
             self._open()
-        try:
-            # TODO: understand why this is necessary
-            self._portal._check_running()
-        except RuntimeError:
-            self._open()
 
         # we have to wrap the method because portal.call doesn't support kwargs
         def wrapped_method():
@@ -925,7 +959,7 @@ class Repo:
             # (e.g. list_prefix(""))
             return [item async for item in func(*args, **kwargs)]
 
-        return self._portal.call(iter_to_list)
+        return self._synchronize(iter_to_list)
 
     def __getstate__(self):
         return self._arepo
@@ -938,10 +972,13 @@ class Repo:
         if self._OPEN:
             try:
                 self._arepo_context.__exit__(None, None, None)
-                self._blocking_portal.__exit__(None, None, None)
             except RuntimeError:
-                # the portal is not running
                 pass
+            self._arepo_context = None
+            self._portal = None
+            self._cached_portal_context.__exit__(None, None, None)
+            self._cached_portal_context = None
+            self._OPEN = False
 
     def __del__(self):
         self.close()
@@ -999,7 +1036,7 @@ class Repo:
             branch_name: New branch name
         """
 
-        return self._arepo.new_branch(branch)
+        return self._synchronize(self._arepo.new_branch, branch)
 
     # TODO: figure out some clever metaclass way of wrapping all of these methods
     # For now it's faster to just plug and chug
@@ -1088,7 +1125,7 @@ class Repo:
 
     @property
     def commit_log(self):
-        return self._arepo.commit_log
+        return self._synchronize(self._arepo.commit_log)
 
     def add_virtual_hdf(self, path: Path, hdf_uri: str) -> None:
         """Add a virtual HDF5 dataset to the arraylake.
